#include <stdio.h>
#include <string.h>
#include <assert.h>

enum ERR
{
   ERR_OK=0,
   ERR_BUFTOOSMALL=-1
};

unsigned short read_word(unsigned char **buf)
{
   unsigned short val=(buf[0][0] << 8) | buf[0][1];
   buf[0] += 2;
   return val;
}

unsigned long read_dword(unsigned char **buf)
{
   unsigned long val=(buf[0][0] << 24) | (buf[0][1] << 16) | (buf[0][2] << 8) | buf[0][3];
   buf[0] += 4;
   return val;
}

void memset_word(void *out, unsigned short val, unsigned long size)
{
   for (unsigned long i = 0; i < size; i++)
      ((unsigned short *)out)[i] = val;
}

void memset_dword(void *out, unsigned long val, unsigned long size)
{
   for (unsigned long i = 0; i < size; i++)
      ((unsigned long *)out)[i] = val;
}

unsigned long get_stream_size(unsigned char **buf)
{
   unsigned short header=read_word(buf);

   if (header & 0x8)
   {
      // Dword Decompressed Size            
      buf[0] += 2; // adjust pointer for alignment
      return read_dword(buf);
   }
   else
   {
      // Word Decompressed Size
      return read_word(buf);
   }
}


int cmp_decompress_byte_stream(unsigned char *buf, unsigned char **out, unsigned long out_size, unsigned long *dcmp_size)
{
   int ret = ERR_OK;
   dcmp_size[0] = get_stream_size(&buf);

   if (dcmp_size[0] > out_size)
      return ERR_BUFTOOSMALL;

   int run_len;
   for (unsigned long i = 0; i < dcmp_size[0]; i += run_len)
   {
      run_len = (char)buf[0];
      buf++;

      if (run_len >= 0)
      {
         // Run
         run_len += 2;

         if (i + run_len > out_size)
            run_len = out_size - i;

         unsigned char data=buf[0];
         buf++;

         memset(out[0], data, run_len);
      }
      else
      {
         // Direct byte copy
         run_len *= -1;

         if (i + run_len > out_size)
            run_len = out_size - i;

         memcpy(out[0], buf, run_len);
         buf += run_len;
      }

      out[0] += run_len;
   }

   return ERR_OK;
}

int cmp_decompress_word_stream(unsigned char *buf, unsigned char **out, unsigned long out_size, unsigned long *dcmp_size)
{
   int ret = ERR_OK;
   dcmp_size[0] = get_stream_size(&buf);
   int end_offset=dcmp_size[0] & 0x1;
   dcmp_size[0] = dcmp_size[0] + (dcmp_size[0] % 2);

   if (dcmp_size[0] > out_size)
      return ERR_BUFTOOSMALL;
   
   int run_len;
   for (unsigned long i = 0; i < out_size; i += run_len << 1) 
   {
      run_len=(signed short)read_word(&buf);

      if (run_len >= 0)
      {
         // Run
         run_len += 2;

         if (i + (run_len<<1) > out_size)
            run_len = (out_size - i) >> 1;

         memset_word(out[0], read_word(&buf), run_len);
      }
      else
      {
         // Direct byte copy
         run_len *= -1;

         if (i + (run_len << 1) > out_size)
            run_len = (out_size - i) >> 1;

         memcpy(out[0], buf, run_len << 1);
         buf += run_len << 1;
      }

      out[0] += run_len << 1;
   }
   out[0] -= end_offset;
   return ret;
}

int cmp_decompress_dword_stream(unsigned char *buf, unsigned char **out, unsigned long out_size, unsigned long *dcmp_size)
{
   int ret = ERR_OK;
   dcmp_size[0] = get_stream_size(&buf);
   int end_offset=dcmp_size[0] & 0x3;
   dcmp_size[0] = dcmp_size[0] + (dcmp_size[0] % 4);

   if (dcmp_size[0] > out_size)
      return ERR_BUFTOOSMALL;

   int run_len;
   for (unsigned long i = 0; i < out_size; i += run_len << 2) 
   {
      run_len=(int)read_dword(&buf);

      if (run_len >= 0)
      {
         // Run
         run_len += 2;

         if (i + (run_len<<2) > out_size)
            run_len = (out_size - i) >> 2;

         memset_dword(out[0], read_dword(&buf), run_len);
      }
      else
      {
         // Direct byte copy
         run_len *= -1;

         if (i + (run_len << 2) > out_size)
            run_len = (out_size - i) >> 2;

         memcpy(out[0], buf, run_len << 2);
         buf += run_len << 2;
      }

      out[0] += run_len << 2;
   }

   out[0] -= end_offset;

   return ret;
}

int cmp_decompress(unsigned char *buf, unsigned char **out, unsigned long out_size, unsigned long *dcmp_size)
{
   unsigned short header = read_word(&buf);

   switch ((header >> 10) & 0x3)
   {
      case 0: // Byte
         cmp_decompress_byte_stream(buf, out, out_size, dcmp_size);
         break;
      case 1: // Word
         cmp_decompress_word_stream(buf, out, out_size, dcmp_size);
         break;
      case 2: // Double-word
         cmp_decompress_dword_stream(buf, out, out_size, dcmp_size);
         break;
      default: break;
   }
   return ERR_OK;
}

